#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>



#include "font.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    ui->label_TotalTime->setText("");

    // connections
    QObject::connect(ui->pushButtonFileSelect, SIGNAL(clicked()), SLOT(slot_pushButtonFileSelect_Clicked()));

    // Clear Time
    ui->label_Milliseconds->setText("");

}

void MainWindow::ProcessFont(QString fileName)
{

    QString _text = "";

    //
    // Read in Font Data
    //

    Font *f = new Font();

    // create raw buffer data array
    QFile fontfile_buffer(fileName);
    if(! fontfile_buffer.open(QIODevice::ReadOnly)){
        QMessageBox::information(this,tr("Unable to Open"),fontfile_buffer.errorString());
        return;
    }

    // read raw, unfiltered data
    QByteArray raw_data = fontfile_buffer.readAll();

    // create filtered data for reading header data
    QFile fontfile(fileName);

    if(! fontfile.open(QIODevice::ReadOnly)){
        QMessageBox::information(this,tr("Unable to Open"),fontfile.errorString());
        return;
    }

    //
    // Process Font Data
    //

    // read raw data
    QDataStream in(&fontfile);
    in.setVersion(QDataStream::Qt_5_0);


    // WOFFHeader
    quint32	signature;      // 0x774F4646 'wOFF'
    quint32	flavor;         // The "sfnt version" of the input font.
    quint32	length;         // Total size of the WOFF file.
    quint16	numTables;      // Number of entries in directory of font tables.
    quint16	reserved;       // Reserved; set to zero.
    quint32	totalSfntSize;  // Total size needed for the uncompressed font data, including the sfnt header, directory, and font tables (including padding).
    quint16	majorVersion;   // Major version of the WOFF file.
    quint16	minorVersion;   // Minor version of the WOFF file.
    quint32	metaOffset;     // Offset to metadata block, from beginning of WOFF file.
    quint32	metaLength;     // Length of compressed metadata block.
    quint32	metaOrigLength; // Uncompressed size of metadata block.
    quint32	privOffset;     // Offset to private data block, from beginning of WOFF file.
    quint32	privLength;     // Length of private data block.


    // Process WOFFHeader data
    in >> signature >> flavor >> length >> numTables >> reserved >> totalSfntSize
       >> majorVersion >> minorVersion >> metaOffset >> metaLength >> metaOrigLength
       >> privOffset >> privLength;

    // process extended metadata block
    if(metaLength != 0){

        // create read buffer
        Bytef * _xml_read_buffer = new Bytef[metaLength];

        // loop and copy data to source buffer
        for(int i = 0; i < metaLength; i++){
            _xml_read_buffer[i] = raw_data[metaOffset + i];
        }

        // create uncompressed write buffer
        Bytef _xml_decompress_buffer = new Bytef[metaOrigLength];

        // decompress
        int _xml_uncompress_result = uncompress(_xml_decompress_buffer, &metaOrigLength,
                               (const Bytef *)_xml_read_buffer,
                               (uLong)metaLength);

        f->zerr(_xml_uncompress_result);

        // parse as xml stream
        extended_xml_meta_data.addData(_xml_decompress_buffer);

    }



    // Build array of table entires

    QList<woff_tableEntry *> WOFFEntries;

    for(int i = 0; i < numTables; i++){

        woff_tableEntry * we = new woff_tableEntry;

        in >> we->tag;          // 4-byte sfnt table identifier.
        in >> we->offset;       // Offset to the data, from beginning of WOFF file.
        in >> we->compLength;   // Length of the compressed data, excluding padding.
        in >> we->origLength;   // Length of the uncompressed table, excluding padding.
        in >> we->origChecksum; // Checksum of the uncompressed table.

        we->tagPretty[3] = (char)(we->tag & 0xff);
        we->tagPretty[2] = (char)((we->tag >> 8) & 0xff);
        we->tagPretty[1] = (char)((we->tag >> 16) & 0xff);
        we->tagPretty[0] = (char)((we->tag >> 24) & 0xff);
        we->tagPretty[4] = '\0';

        // alloc destination buffer
        we->data = new Bytef[we->origLength];

        // Decompress if needed
        if(we->compLength != we->origLength){

            // source buffer
            Bytef * _source = new Bytef[we->compLength];

            // copy buffer
            for(int i = 0; i < we->compLength; i++){
                _source[i] = raw_data[we->offset + i];
            }

            int uc = 0;

            uc = uncompress(we->data, &we->origLength,
                                   (const Bytef *)_source,
                                   (uLong)we->compLength);

            f->zerr(uc);

        } else {

            for (uint j = 0; j < we->origLength; j++) {
                we->data[j] = raw_data[we->offset + j];
            }

        }

        WOFFEntries << we;

    }


    QList<NameEntry> NameEntries;

    QListIterator<woff_tableEntry *> WOFFEntry(WOFFEntries);

    while(WOFFEntry.hasNext()){

        woff_tableEntry *entry = WOFFEntry.next();

        if(strcmp(entry->tagPretty, "name") == 0){

            // http://www.microsoft.com/typography/otspec/name.htm

            // Data Access Help.

            //                         01 02 03 04 05 06 07 08
            // ------------------------------------------------------
            // TIP: Let's say we have: 00 00 00 1A 01 3E 00 00
            // In bytes that's:        00    26    62    00
            // Array Index:            1     3     5     7

            // In other words, the index position we use for uint16 is the
            // byte value -1 of the index value we want.

            ushort format = FIX_SHORT(entry->data[0]);  // Format selector (=0).

            ushort count = FIX_SHORT(entry->data[2]); // Number of name records.

            ushort stringOffset = FIX_SHORT(entry->data[4]); // Offset to start of string storage (from start of table).


            ushort byte_offset = 6; // start at 6 bytes in

            // loop through each name record
            for(int i = 0; i < count; i++){

                NameEntry ne;

                ne.platformID = FIX_SHORT(entry->data[byte_offset + 0]);
                ne.platformSpecificID = FIX_SHORT(entry->data[byte_offset + 2]);
                ne.languageID = FIX_SHORT(entry->data[byte_offset + 4]);
                ne.nameID = FIX_SHORT(entry->data[byte_offset + 6]);
                ne.length = FIX_SHORT(entry->data[byte_offset + 8]);
                ne.offset = FIX_SHORT(entry->data[byte_offset + 10]);

                // grab string value
                char * _name = new char[ne.length + 1];

                for(int t = 0; t < ne.length; t++){
                    _name[t] = entry->data[ne.offset + stringOffset + t];
                }

                _name[t] = '\0'; // terminate C-Style string

                ne.name = _name;


                // add to name entries

                NameEntries << ne;

                byte_offset += 12;

            }


        }

    }

    QListIterator<NameEntry> ne(NameEntries);

    while(ne.hasNext()){
        NameEntry n = ne.next();
        _text += n.name; _text += "\r\n";
    }

    ui->plainTextEdit->appendPlainText(_text);

}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::slot_pushButtonFileSelect_Clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Select Font File"),
                                                    QDir::homePath() + "/Desktop",
                                                    tr("WOFF File file (*.woff)"));

    if(fileName != ""){


        uint _ITT = 125;

        //
        // Start Benchmark Timing
        //
        startTime = startTime.currentTime();


        for (int i = 0; i < _ITT; i++){
            this->ProcessFont(fileName);
        }


        //
        // Stop & Tabulate Benchmark
        //

        endTime = endTime.currentTime();

        int elapsedTime = startTime.secsTo(endTime);

        int elapsedTimeMicro = startTime.msecsTo(endTime);

        ui->plainTextEdit->appendPlainText("\n== || Total Process Time Elapsed (seconds): "
                                     + QString::number(elapsedTime));

        ui->plainTextEdit->appendPlainText("== || Total Process Time Elapsed (milliseconds): "
                                         + QString::number(elapsedTimeMicro));

        // Show Elapsed Time
        ui->label_Milliseconds->setText(QString::number(elapsedTimeMicro));


    }

}

void MainWindow::slot_loadFile()
{

}
