﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Diagnostics;

using System.IO;

using Ionic.Zlib;

namespace FontReader
{

    public struct NameEntry
    {
        public UInt16 platformID { get; set; }
        public UInt16 platformSpecificID { get; set; }
        public UInt16 languageID { get; set; }
        public UInt16 nameID { get; set; }
        public UInt16 length { get; set; }
        public UInt16 offset { get; set; }
        public string name { get; set; }
    }

    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();

            labelMilliseconds.Text = "";
        }


        public void ProcessFont(string FileName)
        {

            StringBuilder sb = new StringBuilder();

            // Read in a raw byte copy
            byte[] raw = File.ReadAllBytes(FileName);

            BinaryReader2 br = new BinaryReader2(File.OpenRead(FileName));


            // Read WOFF Header

            UInt32 woff_header_signature;      // 0x774F4646 'wOFF'
            UInt32 woff_header_flavor;         // The "sfnt version" of the input font.
            UInt32 woff_header_length;         // Total size of the WOFF file.
            UInt16 woff_header_numTables;      // Number of entries in directory of font tables.
            UInt16 woff_header_reserved;       // Reserved; set to zero.
            UInt32 woff_header_totalSfntSize;  // Total size needed for the uncompressed font data, including the sfnt header, directory, and font tables (including padding).
            UInt16 woff_header_majorVersion;   // Major version of the WOFF file.
            UInt16 woff_header_minorVersion;   // Minor version of the WOFF file.
            UInt32 woff_header_metaOffset;     // Offset to metadata block, from beginning of WOFF file.
            UInt32 woff_header_metaLength;     // Length of compressed metadata block.
            UInt32 woff_header_metaOrigLength; // Uncompressed size of metadata block.
            UInt32 woff_header_privOffset;     // Offset to private data block, from beginning of WOFF file.
            UInt32 woff_header_privLength;     // Length of private data block.

            // Read uses little endian, but the file is big endian, so convert using a override class.
            // http://stackoverflow.com/questions/8620885/c-sharp-binary-reader-in-big-endian

            woff_header_signature = br.ReadUInt32();
            woff_header_flavor = br.ReadUInt32();
            woff_header_length = br.ReadUInt32();
            woff_header_numTables = br.ReadUInt16();
            woff_header_reserved = br.ReadUInt16();
            woff_header_totalSfntSize = br.ReadUInt32();
            woff_header_majorVersion = br.ReadUInt16();
            woff_header_minorVersion = br.ReadUInt16();
            woff_header_metaOffset = br.ReadUInt32();
            woff_header_metaLength = br.ReadUInt32();
            woff_header_metaOrigLength = br.ReadUInt32();
            woff_header_privOffset = br.ReadUInt32();
            woff_header_privLength = br.ReadUInt32();

            // iterate over WOFF Table Directories

            List<WOFFTableEntry> WOFFEntries = new List<WOFFTableEntry>();

            for (var i = 0; i < woff_header_numTables; i++)
            {
                WOFFTableEntry t = new WOFFTableEntry();

                char[] _tag = new char[4];
                _tag[0] = br.ReadChar();
                _tag[1] = br.ReadChar();
                _tag[2] = br.ReadChar();
                _tag[3] = br.ReadChar();
                t.tag = new string(_tag);

                t.offset = br.ReadUInt32();
                t.compLength = br.ReadUInt32();
                t.origLength = br.ReadUInt32();
                t.origChecksum = br.ReadUInt32();
                t.data = new byte[t.origLength];

                // grab data block

                // Decompress if needed
                if (t.origLength != t.compLength)
                {

                    byte[] compressed = new byte[t.compLength];
                    byte[] uncompressed = new byte[t.origLength];

                    // copy buffer 
                    for (int j = 0; j < t.compLength; j++)
                    {
                        compressed[j] = raw[t.offset + j];
                    }

                    bool expectRfc1950Header = true;

                    using (MemoryStream ms = new MemoryStream())
                    {
                        ZlibCodec compressor = new ZlibCodec();
                        compressor.InitializeInflate(expectRfc1950Header);

                        compressor.InputBuffer = compressed;
                        compressor.AvailableBytesIn = compressed.Length;
                        compressor.NextIn = 0;
                        compressor.OutputBuffer = uncompressed;

                        foreach (var f in new FlushType[] { FlushType.None, FlushType.Finish })
                        {
                            int bytesToWrite = 0;
                            do
                            {
                                compressor.AvailableBytesOut = Convert.ToInt32(t.origLength);
                                compressor.NextOut = 0;
                                compressor.Inflate(f);

                                bytesToWrite = Convert.ToInt32(t.origLength) - compressor.AvailableBytesOut;
                                if (bytesToWrite > 0)
                                    ms.Write(uncompressed, 0, bytesToWrite);
                            }
                            while ((f == FlushType.None && (compressor.AvailableBytesIn != 0 || compressor.AvailableBytesOut == 0)) ||
                                   (f == FlushType.Finish && bytesToWrite != 0));
                        }

                        compressor.EndInflate();

                        t.data = ms.ToArray();
                    }

                }
                else
                {
                    for (int j = 0; j < t.origLength; j++)
                    {
                        t.data[j] = raw[t.offset + j];
                    }
                }

                WOFFEntries.Add(t);
            }


            // name entries list
            List<NameEntry> NameEntries = new List<NameEntry>();


            // loop through each entry

            foreach (var entry in WOFFEntries)
            {
                if (entry.tag == "name")
                {

                    byte[] _tmp = { entry.data[1], entry.data[0] };
                    UInt16 format = BitConverter.ToUInt16(_tmp, 0);

                    _tmp[0] = entry.data[3]; _tmp[1] = entry.data[2];
                    UInt16 count = BitConverter.ToUInt16(_tmp, 0);

                    // swap byte order
                    _tmp[0] = entry.data[5]; _tmp[1] = entry.data[4];
                    UInt16 stringOffset = BitConverter.ToUInt16(_tmp, 0);


                    int byte_offset = 7;

                    for (int j = 0; j < count; j++)
                    {
                        NameEntry ne = new NameEntry();

                        ne.platformID = entry.data[byte_offset + 0];
                        ne.platformSpecificID = entry.data[byte_offset + 2];
                        ne.languageID = entry.data[byte_offset + 4];
                        ne.nameID = entry.data[byte_offset + 6];

                        // Swap bytes
                        _tmp[0] = entry.data[byte_offset + 8]; _tmp[1] = entry.data[byte_offset + 7];
                        ne.length = BitConverter.ToUInt16(_tmp, 0);

                        // Swap bytes
                        _tmp[0] = entry.data[byte_offset + 10]; _tmp[1] = entry.data[byte_offset + 9];
                        ne.offset = BitConverter.ToUInt16(_tmp, 0);

                        // grab string data

                        byte[] _name = new byte[ne.length];

                        for (int t = 0; t < ne.length; t++)
                        {
                            _name[t] = entry.data[ne.offset + stringOffset + t];
                        }

                        ne.name = Encoding.ASCII.GetString(_name);


                        // add to name entries

                        NameEntries.Add(ne);

                        byte_offset += 12;
                    }

                }
            }


            // Display name Entries
            foreach (NameEntry ne in NameEntries)
            {
                sb.Append(ne.name + "\r\n");
            }

            // update text box with chunks of text, appending in NameEntry loop is very slow. 
            textBox1.Text += sb.ToString();

        }

        OpenFileDialog ofd = new OpenFileDialog();


        private void button1_Click(object sender, EventArgs e)
        {

            ofd.DefaultExt = "woff";
            ofd.Filter = "woff Files (*.woff)|*.woff";
            ofd.ShowDialog();

            if (ofd.FileName != "")
            {

                uint _ITT = 125;

                //
                // Start Benchmark Timing
                //

                Stopwatch stopWatch = new Stopwatch();
                stopWatch.Start();

                for (int i = 0; i < _ITT; i++)
                {
                    ProcessFont(ofd.FileName);
                }


                //
                // Stop & Tabulate Benchmark
                //

                stopWatch.Stop();

                TimeSpan ts = stopWatch.Elapsed;

                string elapsedTime = String.Format("{0:00}:{1:00}:{2:00}.{3:00}",
                    ts.Hours, ts.Minutes, ts.Seconds,
                    ts.Milliseconds / 10);

                // Print Results To Window.
                textBox1.Text += "\n== || Total Process Time Elapsed (seconds): "
                                     + ts.Seconds + "\r\n";

                textBox1.Text += "== || Total Process Time Elapsed (milliseconds): "
                                         + ts.Milliseconds;

                // Print Results To Display Elements
                labelMilliseconds.Text = elapsedTime;

            }


        }

        private void label2_Click(object sender, EventArgs e)
        {

        }

    }
}
